import SimpleITK as sitk
import CircuitSeeker.utility as ut


def apply_transform(
    fix, mov,
    fix_spacing, mov_spacing,
    transform_list,
    transform_spacing=None,
    fix_origin=None,
    mov_origin=None,
    ):
    """
    """

    # convert images to sitk objects
    dtype = fix.dtype
    fix = ut.numpy_to_sitk(fix, fix_spacing, fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, mov_origin)

    # default transform spacing is fixed voxel spacing
    if transform_spacing is None:
        transform_spacing = fix_spacing

    # construct transform
    transform = sitk.CompositeTransform(3)
    for t in transform_list:
        if len(t.shape) == 2:
            t = ut.matrix_to_affine_transform(t)
        elif len(t.shape) == 4:
            t = ut.field_to_displacement_field_transform(t, transform_spacing)
        transform.AddTransform(t)

    # set up resampler object
    resampler = sitk.ResampleImageFilter()
    resampler.SetReferenceImage(sitk.Cast(fix, sitk.sitkFloat32))
    resampler.SetInterpolator(sitk.sitkLinear)
    resampler.SetDefaultPixelValue(0)
    resampler.SetTransform(transform)

    # execute, return as numpy array
    resampled = resampler.Execute(sitk.Cast(mov, sitk.sitkFloat32))
    return sitk.GetArrayFromImage(resampled).astype(dtype)





import numpy as np
import os
import ClusterWrap
import CircuitSeeker.fileio as csio
import CircuitSeeker.utility as ut
from CircuitSeeker.align import affine_align
from CircuitSeeker.transform import apply_transform
import dask.array as da
import dask.delayed as delayed
from scipy.ndimage import median_filter
from scipy.ndimage import gaussian_filter1d
from scipy.ndimage import map_coordinates
from scipy.ndimage import zoom
import zarr
from numcodecs import Blosc
from pathlib import Path
from glob import glob


def distributed_image_mean(
    frames,
    cluster_kwargs={},
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # hdf5 files use dask.array
        if csio.testPathExtensionForHDF5(frames['suffix']):
            frames = csio.daskArrayBackedByHDF5(
                frames['folder'], frames['prefix'],
                frames['suffix'], frames['dataset_path'],
            )
            frames_mean = frames.mean(axis=0, dtype=np.float32).compute()
            frames_mean = np.round(frames_mean).astype(frames[0].dtype)
        # other types use dask.bag
        else:
            frames = csio.daskBagOfFilePaths(
                frames['folder'], frames['prefix'], frames['suffix'],
            )
            nframes = frames.npartitions
            frames_mean = frames.map(csio.readImage).reduction(sum, sum).compute()
            dtype = frames_mean.dtype
            frames_mean = np.round(frames_mean/np.float(nframes)).astype(dtype)

    # return reference to mean image
    return frames_mean


def motion_correct(
    fix, frames,
    fix_spacing, frames_spacing,
    time_stride=1,
    sigma=7,
    cluster_kwargs={},
    **kwargs,
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # wrap fixed data as delayed object
        fix_d = delayed(fix)

        # get total number of frames
        total_frames = len(csio.globPaths(
            frames['folder'], frames['prefix'], frames['suffix'],
        ))

        # create dask array of all frames
        frames_data = csio.daskArrayBackedByHDF5(
            frames['folder'], frames['prefix'],
            frames['suffix'], frames['dataset_path'],
            stride=time_stride,
        )
        compute_frames = frames_data.shape[0]

        # set alignment defaults
        alignment_defaults = {
            'rigid':True,
            'alignment_spacing':2.0,
            'metric':'MS',
            'sampling':'random',
            'sampling_percentage':0.1,
            'optimizer':'RGD',
            'iterations':50,
            'max_step':2.0,
        }
        for k, v in alignment_defaults.items():
            if k not in kwargs:
                kwargs[k] = v
 
        # wrap align function
        def wrapped_affine_align(mov, fix_d):
            mov = mov.squeeze()
            t = affine_align(
                fix_d, mov, fix_spacing, frames_spacing,
                **kwargs,
            )
            e = ut.matrix_to_euler_transform(t)
            p = ut.euler_transform_to_parameters(e)
            return p[None, :]

        params = da.map_blocks(
            wrapped_affine_align, frames_data,
            fix_d=fix_d,
            dtype=np.float64,
            drop_axis=[2, 3,],
            chunks=[1, 6],
        ).compute()

    # (weak) outlier removal and smoothing
    params = median_filter(params, footprint=np.ones((3,1)))
    params = gaussian_filter1d(params, sigma, axis=0)

    # interpolate
    if time_stride > 1:
        x = np.linspace(0, compute_frames-1, total_frames)
        coords = np.meshgrid(x, np.mgrid[:6], indexing='ij')
        params = map_coordinates(params, coords)

    # convert to matrices
    transforms = np.empty((total_frames, 4, 4))
    for i in range(params.shape[0]):
        e = ut.parameters_to_euler_transform(params[i])
        t = ut.affine_transform_to_matrix(e)
        transforms[i] = t

    # return all transforms
    return transforms


def save_transforms(transforms, folder, prefix='rigid_transform'):
    """
    """

    Path(folder).mkdir(parents=True, exist_ok=True)
    for i in range(transforms.shape[0]):
        basename = prefix + f'_frame_{i:08d}.mat'
        path = os.path.join(folder, basename)
        np.savetxt(path, transforms[i])


def read_transforms(folder, prefix='rigid_transform'):
    """
    """

    tp = sorted(glob(folder + '/' + prefix + "*"))
    transforms = np.empty( (len(tp),) + (4,4) )
    for i in range(transforms.shape[0]):
        transforms[i] = np.loadtxt(tp[i])
    return transforms


def resample_frames(
    frames,
    frames_spacing,
    transforms,
    write_path,
    mask=None,
    time_stride=1,
    compression_level=4,
    cluster_kwargs={},
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # create dask array of all frames
        frames_data = csio.daskArrayBackedByHDF5(
            frames['folder'], frames['prefix'],
            frames['suffix'], frames['dataset_path'],
            stride=time_stride,
        )
        total_frames = frames_data.shape[0]

        # wrap transforms as dask array
        # extra dimension to match frames_data ndims
        transforms = transforms[::time_stride, None, :, :]
        transforms_d = da.from_array(transforms, chunks=(1, 1, 4, 4))

        # wrap mask
        mask_d = None
        if mask is not None:
            mask_sh, frame_sh = mask.shape, frames_data.shape[1:]
            if mask_sh != frame_sh:
                mask = zoom(mask, np.array(frame_sh) / mask_sh, order=0)
            mask_d = delayed(mask)

        # wrap transform function
        def wrapped_apply_transform(mov, t, mask_d=None):
            mov = mov.squeeze()
            t = t.squeeze()
            aligned = apply_transform(
                mov, mov, frames_spacing, frames_spacing,
                transform_list=[t,],
            )
            if mask_d is not None:
                aligned = aligned * mask_d
            return aligned[None, ...]

        # apply transform to all frames
        frames_aligned = da.map_blocks(
            wrapped_apply_transform, frames_data, transforms_d,
            mask_d=mask_d,
            dtype=np.uint16,
            chunks=[1,] + list(frames_data.shape[1:]),
        )

        # write in parallel as 4D array to zarr file
        compressor = Blosc(
            cname='zstd',
            clevel=compression_level,
            shuffle=Blosc.BITSHUFFLE,
        )
        aligned_disk = zarr.open(
            write_path, 'w',
            shape=frames_aligned.shape,
            chunks=[1,] + list(frames_data.shape[1:]),
            dtype=frames_aligned.dtype,
            compressor=compressor
        )
        da.to_zarr(frames_aligned, aligned_disk)

        # return reference to zarr store
        return aligned_disk




import numpy as np
from os.path import abspath, dirname, join

package_dir = dirname(abspath(__file__))
networks_dir = join(package_dir, 'networks')
data_dir = join(package_dir, 'data')


# include info useful for matching pretrained network to new data
STARDIST_CONFOCAL_NUCLEARGCAMP = {
    'network_path':
        join(networks_dir, 'stardist_confocal_nucleargcamp'),
    'training_image_path':
        [join(data_dir, 'confocal', 'training_forebrain_crop.nrrd'),
         join(data_dir, 'confocal', 'training_hindbrain_crop.nrrd'),
        ],
    'training_labels_path':
        [join(data_dir, 'confocal', 'training_forebrain_crop_labels.nrrd'),
         join(data_dir, 'confocal', 'training_hindbrain_crop_labels.nrrd'),
        ],
    'training_image_orientation':
        'xyz',
    'training_image_voxel_spacing':
        np.array([0.5189, 0.5189, 1.0]),
    'training_image_histogram_path':
        join(data_dir, 'confocal', 'training_image_histogram.npy'),
    'norm_range':
        np.array([1, 99.8]),
}


def getNetworkSpecs(network):
    """
    """
    if network == 'confocal_nucleargcamp':
        return STARDIST_CONFOCAL_NUCLEARGCAMP




import SimpleITK as sitk
import h5py
from glob import glob
from os import path
import numpy as np
import ClusterWrap

import dask.bag as db
import dask.array as da


def testPathExtensionForHDF5(image_path):
    """
    Returns true if `image_path` has an hdf5 like extension
    Currently: `['.h5', '.hdf5']
    """

    hdf5_extensions = ['.h5', '.hdf5']
    if image_path in hdf5_extensions:
        return True
    else:
        ext = path.splitext(image_path)[-1]
        return ext in hdf5_extensions


def globPaths(folder, prefix, suffix):
    """
    Returns sorted list of all absolue paths matching `folder/prefix*suffix`
    If no such paths are found, throws AsserionError
    """

    folder = path.abspath(folder)
    images = sorted(glob(path.join(folder, prefix) + '*' + suffix))
    assert (len(images) > 0), f"No images in {folder} matching {prefix}*{suffix}"
    return images


def daskBagOfFilePaths(folder, prefix, suffix, npartitions=None):
    """
    Returns dask.bag of absolute paths matching `folder/prefix*suffix`
    Specify bag partitions with `npartitions`; default None
    if `npartitions == None` then each filepath is its own partition
    """

    images = globPaths(folder, prefix, suffix)
    if npartitions is None: npartitions = len(images)
    return db.from_sequence(images, npartitions=npartitions)


def daskArrayBackedByHDF5(folder, prefix, suffix, dataset_path, stride=None):
    """
    Returns dask.array backed by HDF5 files matching absolute path `folder/prefix*suffix`
    You must specify hdf5 dataset path with `dataset_path`
    """

    error_message = "daskArrayBackedByHDF5 requires hdf5 files with .h5 or .hdf5 extension"
    assert (testPathExtensionForHDF5(suffix)), error_message
    images = globPaths(folder, prefix, suffix)
    if stride is not None:
        images = images[::stride]
    dsets = [readHDF5(image, dataset_path) for image in images]
    arrays = [da.from_array(dset, chunks=dset.shape) for dset in dsets]
    return da.stack(arrays, axis=0)


def readHDF5(image_path, dataset_path):
    """
    Returns (lazy) h5py array to dataset at `image_path[dataset_path]`
    """

    error_message = "readHDF5 requires hdf5 files with .h5 or .hdf5 extension"
    assert (testPathExtensionForHDF5(image_path)), error_message
    return h5py.File(image_path, 'r')[dataset_path]


def readImage(image_path, dataset_path=None):
    """
    Returns array like object for dataset at `image_path`
    If `image_path` is an hdf5 file, you must specify `dataset_path`
    If `image_path` is an hdf5 file, the return array is a lazy h5py File object
    Else, `image_path` is read by SimpleITK and an in memory numpy array is returned
    File format must be either hdf5 or supported by SimpleITK file readers
    """

    if testPathExtensionForHDF5(image_path):
        assert(dataset_path is not None), "Must provide dataset_path for .h5/.hdf5 files"
        return readHDF5(image_path, dataset_path)
    else:
        return sitk.GetArrayFromImage(sitk.ReadImage(image_path))


def writeHDF5(image_path, dataset_path, array):
    """
    Writes `array` to `image_path[dataset_path]` as an hdf5 file
    """

    with h5py.File(image_path, 'w') as f:
        dset = f.create_dataset(dataset_path, array.shape, array.dtype)
        dset[...] = array


def writeImage(image_path, array, spacing=None, axis_order='zyx'):
    """
    Writes `array` to `image_path`
    `image_path` extension determines format - must be supported by SimpleITK image writers
    Many formats support voxel spacing in the metadata, to specify set `spacing`
    If format supports voxel spacing in meta data, you can set with `spacing`
    Use `axis_order` to specify current axis ordering; default: 'zyx'
    """

    # argsort axis_order
    transpose_order = [i for (v, i) in sorted((v, i) for (i, v) in enumerate(axis_order))]
    array = array.transpose(transpose_order[::-1])  # sitk inverts axis order
    img = sitk.GetImageFromArray(array)
    if spacing is not None:
        spacing = spacing[transpose_order]
        img.SetSpacing(spacing[::-1])
    sitk.WriteImage(img, image_path)


def ensureArray(reference, dataset_path):
    """
    """

    if not isinstance(reference, np.ndarray):
        if not isinstance(reference, str):
            raise ValueError("image references must be ndarrays or filepaths")
        reference = readImage(reference, dataset_path)[...]  # hdf5 arrays are lazy
    return reference


def stack_to_hdf5(stack_path, write_path, dims, dtype):
    """
    """

    stack = np.fromfile(stack_path, dtype=dtype).reshape(dims)
    writeHDF5(write_path, '/default', stack)


def distributed_stack_to_hdf5(
    folder, prefix, suffix,
    dims, dtype,
    cluster_kwargs={}
):
    """
    """

    stack_paths = globPaths(folder, prefix, suffix)
    write_paths = [path.splitext(s)[0] + '.h5' for s in stack_paths]
    nimages = len(stack_paths)
    stack_paths_b = db.from_sequence(stack_paths, npartitions=nimages)
    write_paths_b = db.from_sequence(write_paths, npartitions=nimages)
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:
        stack_paths_b.map(stack_to_hdf5, write_paths_b, dims, dtype).compute()




import os, sys
import numpy as np
import dask.array as da
import SimpleITK as sitk
import ClusterWrap
import CircuitSeeker.utility as ut
import greedypy.greedypy_registration_method as grm
from scipy.ndimage import minimum_filter, gaussian_filter
from dask_stitch.local_affine import local_affines_to_field


def configure_irm(
    metric='MI', bins=128,
    sampling='regular', sampling_percentage=1.0,
    optimizer='GD', iterations=200, learning_rate=1.0,
    min_step=0.1, max_step=1.0,
    shrink_factors=[2,1],
    smooth_sigmas=[2,1],
    num_steps=[2, 2, 2],
    step_sizes=[1., 1., 1.],
    callback=None,
):
    """
    """

    # set up registration object
    ncores = int(os.environ["LSB_DJOB_NUMPROC"])  # TODO: LSF specific!
    sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(2*ncores)
    irm = sitk.ImageRegistrationMethod()
    irm.SetNumberOfThreads(2*ncores)

    # set interpolator
    irm.SetInterpolator(sitk.sitkLinear)

    # set metric
    if metric == 'MI':
        irm.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=bins,
        )
    elif metric == 'CC':
        irm.SetMetricAsCorrelation()
    elif metric == 'MS':
        irm.SetMetricAsMeanSquares()

    # set metric sampling
    if sampling == 'regular':
        irm.SetMetricSamplingStrategy(irm.REGULAR)
    elif sampling == 'random':
        irm.SetMetricSamplingStrategy(irm.RANDOM)
    irm.SetMetricSamplingPercentage(sampling_percentage)

    # set optimizer
    if optimizer == 'GD':
        irm.SetOptimizerAsGradientDescent(
            numberOfIterations=iterations,
            learningRate=learning_rate,
        )
        irm.SetOptimizerScalesFromPhysicalShift()
    elif optimizer == 'RGD':
        irm.SetOptimizerAsRegularStepGradientDescent(
            minStep=min_step, learningRate=learning_rate,
            numberOfIterations=iterations,
            maximumStepSizeInPhysicalUnits=max_step,
        )
        irm.SetOptimizerScalesFromPhysicalShift()
    elif optimizer == 'EX':
        irm.SetOptimizerAsExhaustive(num_steps[::-1])
        irm.SetOptimizerScales(step_sizes[::-1])

    # set pyramid
    irm.SetShrinkFactorsPerLevel(shrinkFactors=shrink_factors)
    irm.SetSmoothingSigmasPerLevel(smoothingSigmas=smooth_sigmas)
    irm.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    # set callback function
    if callback is None:
        def callback(irm):
            level = irm.GetCurrentLevel()
            iteration = irm.GetOptimizerIteration()
            metric = irm.GetMetricValue()
            print("LEVEL: ", level, " ITERATION: ", iteration, " METRIC: ", metric)
    irm.AddCommand(sitk.sitkIterationEvent, lambda: callback(irm))

    # return configured irm
    return irm


def affine_align(
    fix, mov,
    fix_spacing, mov_spacing,
    rigid=False,
    initial_transform=None,
    initialize_with_centering=False,
    alignment_spacing=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    default=np.eye(4),
    **kwargs,
):
    """
    """

    # update default if rigid is provided
    if initial_transform is not None and np.all(default == np.eye(4)):
        default = initial_transform

    # if using masks, ensure there is sufficient foreground
    FOREGROUND_PERCENTAGE_THRESHOLD = 0.2
    if fix_mask is not None:
        foreground_percentage = np.sum(fix_mask) / np.prod(fix_mask.shape)
        if foreground_percentage < FOREGROUND_PERCENTAGE_THRESHOLD:
            print("Too little foreground data in fixed image")
            print("Returning default")
            sys.stdout.flush()
            return default
    if mov_mask is not None:
        foreground_percentage = np.sum(mov_mask) / np.prod(mov_mask.shape)
        if foreground_percentage < FOREGROUND_PERCENTAGE_THRESHOLD:
            print("Too little foreground data in moving image")
            print("Returning default")
            sys.stdout.flush()
            return default

    # skip sample to alignment spacing
    if alignment_spacing is not None:
        fix, fix_spacing_ss = ut.skip_sample(fix, fix_spacing, alignment_spacing)
        mov, mov_spacing_ss = ut.skip_sample(mov, mov_spacing, alignment_spacing)
        if fix_mask is not None:
            fix_mask, _ = ut.skip_sample(fix_mask, fix_vox, alignment_spacing)
        if mov_mask is not None:
            mov_mask, _ = ut.skip_sample(mov_mask, mov_vox, alignment_spacing)
        fix_spacing = fix_spacing_ss
        mov_spacing = mov_spacing_ss

    # convert to sitk images
    fix = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)

    # set up registration object
    irm = configure_irm(**kwargs)

    # select initial transform type
    if rigid and initial_transform is None:
        transform = sitk.Euler3DTransform()
    elif rigid and initial_transform is not None:
        transform = ut.matrix_to_euler_transform(initial_transform)
    elif not rigid and initial_transform is None:
        transform = sitk.AffineTransform(3)
    elif not rigid and initial_transform is not None:
        transform = ut.matrix_to_affine_transform(initial_transform)

    # consider initializing with centering
    if initial_transform is None and initialize_with_centering:
        transform = sitk.CenteredTransformInitializer(
            fix, mov, transform,
        )

    # set initial transform
    irm.SetInitialTransform(transform, inPlace=True)

    # set masks
    if fix_mask is not None:
        fix_mask = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
        irm.SetMetricFixedMask(fix_mask)
    if mov_mask is not None:
        mov_mask = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
        irm.SetMetricMovingMask(mov_mask)

    # execute alignment
    irm.Execute(
        sitk.Cast(fix, sitk.sitkFloat32),
        sitk.Cast(mov, sitk.sitkFloat32),
    )

    # if centered, convert back to Euler3DTransform object
    if not isinstance(rigid, np.ndarray) and initialize_with_centering:
        transform = sitk.Euler3DTransform(transform)

    # get initial and final metric values
    initial_metric_value = irm.MetricEvaluate(
        sitk.Cast(fix, sitk.sitkFloat32),
        sitk.Cast(mov, sitk.sitkFloat32),
    )
    final_metric_value = irm.GetMetricValue()

    # if registration improved metric return result
    # otherwise return default
    if final_metric_value < initial_metric_value:
        sys.stdout.flush()
        return ut.affine_transform_to_matrix(transform)
    else:
        print("Optimization failed to improve metric")
        print("Returning default")
        sys.stdout.flush()
        return default


def piecewise_affine_align(
    fix, mov,
    fix_spacing, mov_spacing,
    nblocks,
    pad=16,
    fix_mask=None,
    mov_mask=None,
    cluster_kwargs={},
    **kwargs,
):
    """
    """

    # get default masks
    if fix_mask is None:
        fix_mask = np.ones(fix.shape, dtype=np.uint8)
    if mov_mask is None:
        mov_mask = np.ones(mov.shape, dtype=np.uint8) 

    # compute block size and overlaps
    blocksize = np.array(fix.shape).astype(np.float32) / nblocks
    blocksize = np.ceil(blocksize).astype(np.int16)
    overlaps = blocksize // 2

    # set up cluster
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # construct dask array versions of objects
        # extra dimensions to match field of affine matrices
        fix_da = da.from_array(fix)
        mov_da = da.from_array(mov)
        fm_da = da.from_array(fix_mask)
        mm_da = da.from_array(mov_mask)

        # pad the ends to fill in the last blocks
        # blocks must all be exact for stitch to work correctly
        pads = [(0, y - x % y) if x % y > 0
            else (0, 0) for x, y in zip(fix.shape, blocksize)]
        fix_da = da.pad(fix_da, pads).rechunk(tuple(blocksize))
        mov_da = da.pad(mov_da, pads).rechunk(tuple(blocksize))
        fm_da = da.pad(fm_da, pads).rechunk(tuple(blocksize))
        mm_da = da.pad(mm_da, pads).rechunk(tuple(blocksize))

        # closure for affine alignment
        def single_affine_align(fix, mov, fm, mm, block_info=None):
            # rigid alignment
            rigid = affine_align(
                fix, mov, fix_spacing, mov_spacing,
                fix_mask=fm, mov_mask=mm,
                rigid=True,
                **kwargs,
            )
            # affine alignment
            affine = affine_align(
                fix, mov, fix_spacing, mov_spacing,
                fix_mask=fm, mov_mask=mm,
                initial_transform=rigid,
                **kwargs,
            )
            # correct for block origin
            idx = block_info[0]['chunk-location']
            origin = np.maximum(0, blocksize * idx - overlaps)
            origin = origin * fix_spacing
            tl, tr = np.eye(4), np.eye(4)
            tl[:3, -1], tr[:3, -1] = origin, -origin
            affine = np.matmul(tl, np.matmul(affine, tr))
            # return result
            return affine.reshape((1,1,1,4,4))

        # affine align all chunks
        affines = da.map_overlap(
            single_affine_align,
            fix_da, mov_da, fm_da, mm_da,
            depth=tuple(overlaps),
            dtype=np.float32,
            boundary='none',
            trim=False,
            align_arrays=False,
            new_axis=[3,4],
            chunks=[1,1,1,4,4],
        ).compute()

        # stitch local affines into displacement field
        field = local_affines_to_field(
            fix.shape, fix_spacing,
            affines, blocksize, overlaps,
        ).compute()

        return field


def exhaustive_translation(
    fix, mov,
    fix_spacing, mov_spacing,
    num_steps, step_sizes,
    fix_origin=None,
    mov_origin=None,
    peak_ratio=1.2,
    **kwargs,
):
    """
    """

    # squeeze any negligible dimensions
    fix = fix.squeeze()
    mov = mov.squeeze()

    # convert to sitk images
    fix_itk = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
    mov_itk = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)

    # define callback: keep track of alignment scores
    scores = np.zeros(tuple(2*x+1 for x in num_steps[::-1]), dtype=np.float32)
    def callback(irm):
        iteration = irm.GetOptimizerIteration()
        indx = np.unravel_index(iteration, scores.shape, order='F')
        scores[indx[0], indx[1], indx[2]] = irm.GetMetricValue()

    # get irm
    kwargs['optimizer'] = 'EX'
    kwargs['num_steps'] = num_steps
    kwargs['step_sizes'] = step_sizes * fix_spacing
    kwargs['callback'] = callback
    irm = configure_irm(**kwargs)

    # set translation transform
    irm.SetInitialTransform(sitk.TranslationTransform(3), inPlace=True)

    # align
    irm.Execute(
        sitk.Cast(fix_itk, sitk.sitkFloat32),
        sitk.Cast(mov_itk, sitk.sitkFloat32),
    )

    # get best two local minima
    peaks = (minimum_filter(scores, size=3) == scores)
    scores[~peaks] = 0
    min1_indx = np.unravel_index(np.argmin(scores), scores.shape)
    min1 = scores[min1_indx[0], min1_indx[1], min1_indx[2]]
    scores[min1_indx[0], min1_indx[1], min1_indx[2]] = 0
    min2_indx = np.unravel_index(np.argmin(scores), scores.shape)
    min2 = scores[min2_indx[0], min2_indx[1], min2_indx[2]]

    # determine if minimum is good enough
    trans = np.zeros(3)
    if min1 <= min2*peak_ratio:
        trans = np.array(min1_indx[::-1]) - num_steps
        trans = trans * step_sizes * fix_spacing

    # return translation in xyz order
    return trans[::-1]


def piecewise_exhaustive_translation(
    fix, mov,
    fix_spacing, mov_spacing,
    stride,
    query_radius,
    num_steps,
    step_sizes,
    smooth_sigma,
    mask=None,
    cluster_kwargs={},
    **kwargs,
    ):
    """
    """

    # compute search radius in voxels
    search_radius = [q+x*y for q, x, y in zip(query_radius, num_steps, step_sizes)]

    # compute edge pad size
    limit = [x if x > y else y for x, y in zip(search_radius, stride)]

    # get valid sample points as coordinates
    samples = np.zeros_like(fix)
    samples[limit[0]:-limit[0]:stride[0],
            limit[1]:-limit[1]:stride[1],
            limit[2]:-limit[2]:stride[2]] = 1
    if mask is not None:
        samples = samples * mask
    samples = np.nonzero(samples)

    # prepare arrays to hold fixed and moving blocks
    nsamples = len(samples[0])
    fix_blocks_shape = (nsamples,) + tuple(x*2 for x in search_radius)
    mov_blocks_shape = (nsamples,) + tuple(x*2 for x in query_radius)
    fix_blocks = np.empty(fix_blocks_shape, dtype=fix.dtype)
    mov_blocks = np.empty(mov_blocks_shape, dtype=mov.dtype)

    # get context for all sample points
    for i, (x, y, z) in enumerate(zip(samples[0], samples[1], samples[2])):
        fix_blocks[i] = fix[x-search_radius[0]:x+search_radius[0],
                            y-search_radius[1]:y+search_radius[1],
                            z-search_radius[2]:z+search_radius[2]]
        mov_blocks[i] = mov[x-query_radius[0]:x+query_radius[0],
                            y-query_radius[1]:y+query_radius[1],
                            z-query_radius[2]:z+query_radius[2]]

    # compute the query_block origin in physical units
    mov_origin = np.array(search_radius) - query_radius
    mov_origin = mov_origin * fix_spacing

    # set up cluster
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # convert to dask arrays
        fix_blocks_da = da.from_array(
            fix_blocks, chunks=(1,)+fix_blocks.shape[1:],
        )
        mov_blocks_da = da.from_array(
            mov_blocks, chunks=(1,)+mov_blocks.shape[1:],
        )

        # closure for exhaustive translation alignment
        def wrapped_exhaustive_translation(x, y):
            t = exhaustive_translation(
                x, y, fix_spacing, mov_spacing,
                num_steps, step_sizes,
                mov_origin=mov_origin,
                **kwargs,
            )
            return np.array(t).reshape((1, 3))

        # distribute
        translations = da.map_blocks(
            wrapped_exhaustive_translation,
            fix_blocks_da, mov_blocks_da,
            dtype=np.float64, 
            drop_axis=[2,3],
            chunks=[1, 3],
        ).compute()

    # reformat to displacement vector field
    dvf = np.zeros(fix.shape + (3,), dtype=np.float32)
    weights = np.pad([[[1.]]], [(s, s) for s in stride], mode='linear_ramp')
    for t, x, y, z in zip(translations, samples[0], samples[1], samples[2]):
        s = [slice(max(0, x-stride[0]), x+stride[0]+1),
             slice(max(0, y-stride[1]), y+stride[1]+1),
             slice(max(0, z-stride[2]), z+stride[2]+1),]
        dvf[tuple(s)] += t * weights[..., None]

    # smooth
    dvf_s = np.empty_like(dvf)
    for i in range(3):
        dvf_s[..., i] = gaussian_filter(dvf[..., i], smooth_sigma/fix_spacing)

    # return
    return dvf_s







# TODO: still need to refactor deformable

def deformable_align_greedypy(
    fixed, moving,
    fixed_vox, moving_vox,
    radius=32,
    gradient_smoothing=[3.0, 0.0, 1.0, 2.0],
    field_smoothing=[0.5, 0.0, 1.0, 6.0],
    iterations=[200,100],
    shrink_factors=[2,1],
    smooth_sigmas=[1,0],
    step=5.0,
    ):
    """
    """

    register = grm.greedypy_registration_method(
        fixed, fixed_vox,
        moving, moving_vox,
        iterations,
        shrink_factors,
        smooth_sigmas,
        radius=radius,
        gradient_abcd=gradient_smoothing,
        field_abcd=field_smoothing,
    )

    register.mask_values(0)
    register.optimize()
    return register.get_warp()


def piecewise_deformable_align_greedypy(
    fixed, moving,
    fixed_vox, moving_vox,
    affine_matrix,
    block_size=[96,96,96],
    overlap=16, 
    distributed_state=None,
    ):
    """
    """

    # reasmple moving image with affine
    moving_res = apply_transform(
        fixed, moving,
        fixed_vox, moving_vox,
        matrix=affine_matrix
    )

    # set up cluster
    # TODO: need way to pass distributed_state as context manager?
    with csd.distributedState() as ds:

        # TODO: expose cores/tpw, remove job_extra -P
        ds.initializeLSFCluster(
            job_extra=["-P scicompsoft"],
            ncpus=1,
            cores=1,
            threads_per_worker=2,
            memory="15GB",
            mem=15000,
        )
        ds.initializeClient()
        nchunks = np.ceil(np.array(fixed.shape)/block_size)
        ds.scaleCluster(njobs=np.prod(nchunks))

    # TODO: refactor into a function, generalize w.r.t. dimension, share on github
    # TODO: pad array so large overlaps will work (chunk can't be smaller than overlap)
    # chunk ndarrays onto workers and stack as single dask array
        bs = block_size  # shorthand
        fixed_blocks = [[
            [da.from_array(fixed[i:i+bs[0], j:j+bs[1], k:k+bs[2]])
            for k in range(0, fixed.shape[2], bs[2])]
            for j in range(0, fixed.shape[1], bs[1])]
            for i in range(0, fixed.shape[0], bs[0])]
        fixed_da = da.block(fixed_blocks)
        moving_blocks = [[
            [da.from_array(moving_res[i:i+bs[0], j:j+bs[1], k:k+bs[2]])
            for k in range(0, moving_res.shape[2], bs[2])]
            for j in range(0, moving_res.shape[1], bs[1])]
            for i in range(0, moving_res.shape[0], bs[0])]
        moving_da = da.block(moving_blocks)
    
        # deform all chunks
        # TODO: need way to get registration parameters as input to this function
        compute_blocks = [x + 2*overlap for x in block_size] + [3,]
        deformation = da.map_overlap(
            lambda w,x,y,z: deformableAlign(w, x, y, z),
            fixed_da, moving_da,
            depth=overlap,
            dtype=np.float32,
            chunks=compute_blocks,
            new_axis=[3,],
            align_arrays=False,
            boundary='reflect',
            y=fixed_vox, z=fixed_vox,
        ).compute()

    # TODO: TEMP
    resampled = apply_transform(
       fixed, moving_res, fixed_vox, fixed_vox, displacement=deformation
    )

    return deformation, resampled


def deformable_align(
    fix, mov,
    fix_spacing, mov_spacing,
    control_point_spacing,
    control_point_levels,
    initial_transform=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    **kwargs,
):
    """
    """

    # convert to sitk images
    fix = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)

    # set up registration object
    irm = configure_irm(**kwargs)

    # set initial moving transform
    if initial_transform is not None:
        if len(initial_transform.shape) == 2:
            it = ut.matrix_to_affine_transform(initial_transform)
        irm.SetMovingInitialTransform(it)

    # get control point grid shape
    fix_size_physical = [sz*sp for sz, sp in zip(fix.GetSize(), fix.GetSpacing())]
    x, y = control_point_spacing, control_point_levels[-1]
    control_point_grid = [max(1, int(sz / (x*y))) for sz in fix_size_physical]

    # set initial transform
    transform = sitk.BSplineTransformInitializer(
        image1=fix, transformDomainMeshSize=control_point_grid, order=3,
    )
    irm.SetInitialTransformAsBSpline(
        transform, inPlace=True, scaleFactors=control_point_levels,
    )

    # set masks
    if fix_mask is not None:
        fix_mask = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
        irm.SetMetricFixedMask(fix_mask)
    if mov_mask is not None:
        mov_mask = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
        irm.SetMetricMovingMask(mov_mask)

    # execute alignment
    irm.Execute(
        sitk.Cast(fix, sitk.sitkFloat32),
        sitk.Cast(mov, sitk.sitkFloat32),
    )

    # get initial and final metric values
    initial_metric_value = irm.MetricEvaluate(
        sitk.Cast(fix, sitk.sitkFloat32),
        sitk.Cast(mov, sitk.sitkFloat32),
    )
    final_metric_value = irm.GetMetricValue()

    # if registration improved metric return result
    # otherwise return default
    if final_metric_value < initial_metric_value:
        sys.stdout.flush()
        return ut.bspline_to_displacement_field(fix, transform)
    else:
        print("Optimization failed to improve metric")
        print("Returning default")
        sys.stdout.flush()
        return None




import CircuitSeeker.fileio as csio
import CircuitSeeker.networks as csn

from stardist.models import StarDist3D
from csbdeep.utils import normalize

import numpy as np
from scipy.ndimage import zoom

def toXYZ(order):
    """
    """

    return [i for (v, i) in sorted((v, i) for (i, v) in enumerate(order))]


def transpose(image, spacing, orientation, target_orientation):
    """
    """

    image_to_xyz = toXYZ(orientation)
    xyz_to_target = np.argsort(toXYZ(target_orientation))
    image_trans = image.transpose(image_to_xyz).transpose(xyz_to_target)
    spacing_trans = spacing[image_to_xyz][xyz_to_target]
    return image_trans.copy(), spacing_trans  # copy ensures c order after transpose


def resample(image, spacing, target_spacing, order=1):
    """
    """

    # zoom image to match target voxel spacing
    return zoom(image, spacing/target_spacing, order=order)


def rescaleIntensity(image, target_histogram,
    winsor_min=0.001, winsor_max=0.999):
    """
    """

    # linearly rescale intensities using Winsorized min/max
    cdf = np.cumsum(target_histogram[0])
    cdf = cdf / cdf[-1]
    new_mn = target_histogram[1, np.argmax(cdf >= winsor_min)]
    new_mx = target_histogram[1, np.argmax(cdf >= winsor_max)-1]
    old_mn, old_mx = np.quantile(image, [winsor_min, winsor_max])
    return (image - old_mn) * (new_mx - new_mn) / (old_mx - old_mn) + new_mn


def segmentPretrainedStardist(image, spacing, orientation,
    model='confocal_nucleargcamp', write_path=None,
    dataset_path=None, n_tiles=(16, 32, 2),
    prob_thresh=0.5, nms_thresh=0.3):
    """
    """

    # get network details
    ns = csn.getNetworkSpecs(model)
    target_orientation = ns['training_image_orientation']
    target_spacing = ns['training_image_voxel_spacing']
    target_hist = np.load(ns['training_image_histogram_path'])
    target_norm = ns['norm_range']
    network = ns['network_path']

    # load image and normalize
    image = csio.ensureArray(image, dataset_path)
    image_res, spacing_res = transpose(
        image, spacing, orientation, target_orientation,
    )
    image_res = resample(image_res, spacing_res, target_spacing)
    image_res = rescaleIntensity(image_res, target_hist)
    image_norm = normalize(image_res, target_norm[0], target_norm[1])

    # segment
    model = StarDist3D(None, name='stardist', basedir=network)
    pred, det = model.predict_instances(image_norm, n_tiles=n_tiles,
        prob_thresh=prob_thresh, nms_thresh=nms_thresh, verbose=True,
        sparse=True, nms_kwargs={'use_kdtree':True},
    )

    # return to native spacing and orientation
    pred_res = resample(pred, target_spacing, spacing_res, order=0)
    pred_res, _ = transpose(
        pred_res, spacing_res, target_orientation, orientation
    )

    # save results
    if write_path is not None:
        csio.writeImage(write_path, pred_res, axis_order=orientation)
    return pred

